from logs.log import Log
from typing import List, Dict
from sessionizator.strategies.session_path import SessionPath


def get_sessions_with_time_referer_heuristic(logs: List[Log]) -> List[SessionPath]:
  if not logs:
    return []
    
  logs.sort(key=lambda log: log.time)
  time_delta = get_delta()

  reference_to_path_node_by_uri: Dict[str, SessionPath] = {}
  sessions: List[SessionPath] = []
  seen = set()
  for log in logs:
    previous_path = reference_to_path_node_by_uri[log.referer_uri] if log.referer_uri in reference_to_path_node_by_uri else None

    if previous_path and log.time - previous_path.last_log_time > time_delta:
      # previous_path has ended and it's a session 
      # it is not required to be done here, it's an optimization
      # for the latter 'get_sessions' function
      to_add = previous_path.to_session(reference_to_path_node_by_uri, seen)
      if to_add:
        sessions.append(to_add.super_parent)
      
      previous_path = None
    
    # This MAY be the case when the user goes "back" on the web page
    # 
    # It should be noted that there won't be any cycles on our reference_to_path_node_by_uri
    # since we are always over-writing whenever we visit an already
    # seen URI.
    # 
    # Also, on our reference_to_path_node_by_uri we are storing the path in a linkedlist-like
    # structure. Every Log has a corresponding node, and thus we won't
    # have cycles either. (If the nodes held a 1-1 relationship with the
    # URIs, then there might be cycles. But that's not the case).
    if log.uri in reference_to_path_node_by_uri: # There was a cycle
      if log.time - reference_to_path_node_by_uri[log.uri].last_log_time > time_delta:
        # will lose any previous data, so it needs to be stored
        # this also means the session has finished
        finished_path = reference_to_path_node_by_uri[log.uri]
        to_add = finished_path.to_session(reference_to_path_node_by_uri, seen)
        if to_add:
          sessions.append(to_add.super_parent)
      # else:
        # within the time delta going back and overwriting
        # the reference_to_path_node_by_uri wont be a problem, because we are essentially
        # updating the last node.
        # Considering what happens when there are ramifications
        # (such as opening many tabs), and we visit a node from another 
        # open branch. We should should only end the ssession if we 
        # have a reference to the last node in the session and it is 
        # already expired

    current_path = SessionPath(log, previous_path)
    reference_to_path_node_by_uri[log.uri] = current_path

  get_remaining_sessions(reference_to_path_node_by_uri, sessions, seen)

  return sessions

def get_remaining_sessions(reference_to_path_node_by_uri: Dict[str, SessionPath], sessions: List[SessionPath], seen: set) -> None:
  for _, path in reference_to_path_node_by_uri.items():
    super_parent = path.super_parent
    if super_parent.hash_id in seen:
      continue

    sessions.append(super_parent)
    seen.add(super_parent.hash_id)


def get_delta() -> float:
  return 600.0 # 10 minutes in seconds